
/*
 Name:		Sketch1.ino
 Created:	28.10.2021 13:24:48
 Author:	Andreas Ebner 
*/




#include <TimeLib.h>
#include <ArduinoJson.hpp>
#include <ArduinoJson.h>
#include <Adafruit_BME280.h>
#include <SD.h>
#include <sdios.h>
#include <SdFatConfig.h>
#include <SdFat.h>
#include <RingBuf.h>
#include <MinimumSerial.h>
#include <FreeStack.h>
#include <BufferedPrint.h>
#include <StackTrace.h>
#include <WifiIPStack.h>
#include <MQTTPacket.h>
#include <MQTTUnsubscribe.h>
#include <MQTTSubscribe.h>
#include <MQTTPublish.h>
#include <MQTTLogging.h>
#include <MQTTFormat.h>
#include <MQTTConnect.h>
#include <MQTTClient.h>
#include <IPStack.h>
#include <FP.h>
#include <Countdown.h>
#include <NativeEthernet.h>
#include <string>
#include "updater.h"

#define SEALEVELPRESSURE_HPA (1013.25)

// Konfiguration 

// logfile nach 8.3 Dateinamenskonvention
String LogFile = "datalog.txt";

// verwendete MAC Adresse der Netzwerkkarte
byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };

// Optionen -- werden durch config.jsn auf der SD-Karte überschrieben, falls vorhanden
String sendtopic = "Sensors/beta01/output";
String receivetopic = "Sensors/beta01/input";
char* hostname = "homeserver";
int port = 1883; // Normalerweise 1883
String mqttUserName = "";
String mqttPassword = "";
String mqttClientID = "Sensor_001";
bool useMQTT = false;
bool useLocalLog = true;
ulong f = 0UL;
// Lokale Objekte
Adafruit_BME280 bme; // I2C Breakout von Watterott für Bosch BME280/BMP280

MQTT::Message message;
EthernetClient c;
IPStack ipstack(c);

//digitale Ausgänge für LED
int NetworkPin = 12;
int BrokerPin = 14;
int SensorPin = 15;
int MeasurePin = 16;

// 128 Chars Payload
MQTT::Client<IPStack, Countdown, 162, 1> client = MQTT::Client<IPStack, Countdown, 162, 1>(ipstack);
//MQTT::Client<IPStack, Countdown, 100, 1> client = MQTT::Client<IPStack, Countdown, 100, 1>(ipstack); // gives 66 bytes of payload
//MQTT::Client<IPStack, Countdown, 50, 1> client = MQTT::Client<IPStack, Countdown, 50, 1>(ipstack);  // gives 16 bytes of payload

bool hasSDCard = false;
bool hasMQTTConnection = false;
bool isSensorPresent = false;
void doMeasure(int runs, int Delay = 5);

/// <summary>
/// Sendet einen String als Nachricht �ber den konfigurierten MQTT-Client
/// </summary>
/// <param name="MsgPayload">String mit der Nachricht</param>
/// <param name="Topic">optional String mit einem Subthema</param>
void SendMQTT(String MsgPayload, String Topic = "")
{
	if (!useMQTT) {
		Serial.println("MQTT switched off.");
		return;
	}
	Serial.print("Pushing to ");

	message.qos = MQTT::QOS0;
	message.retained = false;
	message.dup = false;
	message.payload = (void*)MsgPayload.c_str();
	message.payloadlen = strlen(MsgPayload.c_str());
	if (Topic != "")
	{
		String bfr = sendtopic;
		bfr.append("/");
		bfr.append(Topic);
		Serial.println(bfr);
		int rc = client.publish(bfr.c_str(), message);

	}
	else
	{
		Serial.println(sendtopic);
		int rc = client.publish(sendtopic.c_str(), message);
	}
}

/// <summary>
/// Schreibt einen String ans Ende einer Textdatei, erstellt die Datei bei Bedarf neu
/// </summary>
/// <param name="FileName">String : Dateiname </param>
/// <param name="output">String : Daten, die gespeichert werden.</param>
/// <param name="Append">optional - Boolean : Soll die Datei weitergef�hrt werden? Bei false wird sie �berschrieben (ToDo)</param>
void WriteLogSD(String FileName, String output, boolean Append = true)
{
	if (!useLocalLog) {
		Serial.println("Logging switched off.");
		return;
	}
	if (hasSDCard == true)
	{
		if (!Append)
		{
			SD.remove(FileName.c_str());
			// Todo: Delete file before opening
		}
		File dataFile = SD.open(FileName.c_str(), FILE_WRITE);

		if (dataFile) {

			dataFile.println(output);

			dataFile.close();
			Serial.print("Logged to file: ");
			Serial.println(output);

		}
	}
}

void messageArrived(MQTT::MessageData& md)
{
	MQTT::Message& message = md.message;
	String Content = ((char*)message.payload);
	Content = Content.substring(0, message.payloadlen);

	//MessagePath ist die Weiche für die Funktionalität
	int pathlen = md.topicName.lenstring.len;
	auto MsgPath = md.topicName.lenstring.data;
	String MessagePath = "";
	for (size_t i = 0; i < pathlen; i++)
	{
		MessagePath.append(MsgPath[i]);
	}
	MessagePath = MessagePath.substring(receivetopic.length() - 1);
	Serial.print("Message received at: ");
	Serial.println(MessagePath);
	if (MessagePath == "trigger")
	{
		if (Content == "tt")
		{
			doMeasure(1, 0);
		}
		else if (Content.startsWith("tt,"))
		{
			int Startpunkt = Content.indexOf(',');
			int Mittelpunkt = Content.indexOf(',', Startpunkt + 1);
			String loops = Content.substring(Startpunkt + 1, Mittelpunkt);
			String Delay = Content.substring(Mittelpunkt + 1);
			auto l = loops.toInt();
			auto d = Delay.toInt();
			doMeasure(l, d);
		}

	}
	else if (MessagePath == "system")
	{
		if (Content.startsWith("update") && (hasSDCard == true))
		{
			InitUpdate();
		}
	}



}


void DoUpdate(String HexFilePath)
{
	Serial.print(" here with update");
	update_firmware(HexFilePath);

}


void InitUpdate()
{
	auto updroot = SD.open("/update/", READ_ONLY);
	if (!updroot)
	{
		Serial.println("Update Directory not found");
		return;
	}
	File updatehex = updroot.openNextFile();
	if (!updatehex)
	{
		Serial.println("Update File not found");
		return;
	}
	while (updatehex.isDirectory())
	{
		updatehex = updroot.openNextFile();
	}

	if (!updatehex)
	{
		SendMQTT("Update File not found", "sensors/debug");
		return;
	}
	else
	{
		String bfr = "File found: ";
		bfr.append(updatehex.name());
		String fnwp = "/update/";
		fnwp.append(updatehex.name());
		updatehex.close();
		SendMQTT(bfr.c_str(), "sensors/debug");
		DoUpdate(fnwp);
	}


}

void doMeasure(int runs, int Delay = 5)
{
	//long tm = millis();
	for (int i = 0; i < runs; i++)
	{
		digitalWrite(MeasurePin, HIGH);
		String measure = readValues();
		WriteLogSD(LogFile, measure);
		SendMQTT(measure);
		digitalWrite(MeasurePin, LOW);
		if (Delay > 0)
		{
			delay(Delay);
		}
	}

	return;
}

String readValues() {
	if (f == ULONG_MAX) {
		f = 0;
	}
	f = f + 1;
	float Temp = bme.readTemperature();
	float Pressure = bme.readPressure() / 100.0F;
	float Humidity = bme.readHumidity();

	// als json zurückgeben für mqtt
	return "{\"d\":\"" + CurrentTimeAsString()
		+ "\",\"m\":\"" + String(millis())
		+ "\",\"t\":\"" + String(Temp, 3)
		+ "\",\"p\":\"" + String(Pressure, 3)
		+ "\",\"h\":\"" + String(Humidity, 3)
		+ "\",\"id\":\"" + String(f)
		+ "\"}";
}

void connectMQTT()
{
	Serial.print("Connecting to ");
	Serial.print(hostname);
	Serial.print(":");
	Serial.println(port);

	int rc = ipstack.connect(hostname, port);
	if (rc != 1)
	{
		Serial.print("rc from TCP connect is ");
		Serial.println(rc);
		return;
	}

	Serial.println("MQTT connecting");
	MQTTPacket_connectData data = MQTTPacket_connectData_initializer;
	data.MQTTVersion = 3;
	data.clientID.cstring = mqttClientID.c_str();
	data.username.cstring = mqttUserName.c_str();
	data.password.cstring = mqttPassword.c_str();
	rc = client.connect(data);

	if (rc != 0)
	{
		Serial.print("rc from MQTT connect is ");
		Serial.println(rc);
	}
	Serial.println("MQTT connected");
	rc = client.subscribe(receivetopic.c_str(), MQTT::QOS0, messageArrived);
	if (rc != 0)
	{
		Serial.print("rc from MQTT subscribe is ");
		Serial.println(rc);
	}
	Serial.println("MQTT subscribed");
	hasMQTTConnection = true;
	digitalWrite(BrokerPin, HIGH);
}

void connectSDCard()
{
	if (!SD.begin(BUILTIN_SDCARD)) {
		Serial.println("SD Card not initialized.");
		hasSDCard = false;
	}
	else {
		Serial.print("SD Card initialized. Size is: ");
		Serial.println(SD.totalSize());
		Serial.print("SD Card initialized. Used Size is: ");
		Serial.println(SD.usedSize());
		hasSDCard = true;
	};
}

void connectSensor()
{
	unsigned status;

	// default settings
	status = bme.begin();
	// You can also pass in a Wire library object like &Wire2
	// status = bme.begin(0x76, &Wire2)
	if (!status) {
		isSensorPresent = false;
		Serial.println("sensor not found");
	}
	else {
		isSensorPresent = true;
		Serial.println("sensor exists");
		digitalWrite(SensorPin, HIGH);
	}
}

time_t getTeensyTime()
{
	return Teensy3Clock.get();
}

void TimeSet() {
	setSyncProvider(getTeensyTime);
	if (timeStatus() != timeSet) {
		Serial.println("Unable to sync with the RTC");
	}
	else {
		Serial.println("");
		Serial.println("RTC has set the system time");
		Serial.println(CurrentTimeAsString());
	}
}

String CurrentTimeAsString() {
	String buffer = "";
	time_t snap = now();
	buffer = String(year(snap)) + "-"
		+ ("00" + String(month(snap))).substring(("00" + String(month(snap))).length() - 2) + "-"
		+ ("00" + String(day(snap))).substring(("00" + String(day(snap))).length() - 2) + " "
		+ ("00" + String(hour(snap))).substring(("00" + String(hour(snap))).length() - 2) + ":"
		+ ("00" + String(minute(snap))).substring(("00" + String(minute(snap))).length() - 2) + ":"
		+ ("00" + String(second(snap))).substring(("00" + String(second(snap))).length() - 2);
	return buffer;
}

void readConfig()
{
	File dataFile = SD.open("config.jsn", FILE_READ);
	String Bfr = "";
	if (dataFile) {
		// json file vollständig einlesen
		while (dataFile.available()) {
			Bfr.append(dataFile.readString());
		}
		dataFile.close();


		ARDUINOJSON_NAMESPACE::StaticJsonDocument<1024> config;
		auto errr = ARDUINOJSON_NAMESPACE::deserializeJson(config, Bfr);  // sieht in der IDE wie ein Fehler aus, ist
																		 // aber nur ein fehlender Header für Intellisense
		if (errr != 0) return;

		Serial.println("Config:");

		if (config["sendtopic"]) {
			Serial.println(config["sendtopic"].as<String>());
			sendtopic = config["sendtopic"].as<String>();
		}
		if (config["receivetopic"]) {
			Serial.println(config["receivetopic"].as<String>());
			receivetopic = config["receivetopic"].as<String>();
		}
		if (config["mqtthostname"]) {
			Serial.println(config["mqtthostname"].as<String>());
			//hostname = config["mqtthostname"].as<String>().c_str();
		}
		if (config["port"]) {
			Serial.println((int)config["port"]);
			port = (int)config["port"];
		}
		if (config["mqttUserName"]) {
			Serial.println(config["mqttUserName"].as<String>());
			mqttUserName = config["mqttUserName"].as<String>();
		}
		if (config["mqttPassword"]) {
			Serial.println(config["mqttPassword"].as<String>());
			mqttPassword = config["mqttPassword"].as<String>();
		}
		if (config["mqttClientID"]) {
			Serial.println(config["mqttClientID"].as<String>());
			mqttClientID = config["mqttClientID"].as<String>();
		}
		if (config["useMQTT"]) {
			useMQTT = true;
		}
		else
		{
			useMQTT = false;
		}
		if (config["useLocalLog"]) {

			useLocalLog = true;
		}
		else
		{
			useLocalLog = false;
		}
	}




}


void setup() {

	pinMode(MeasurePin, OUTPUT);
	digitalWrite(MeasurePin, HIGH);
	pinMode(NetworkPin, OUTPUT);
	digitalWrite(NetworkPin, LOW);
	pinMode(BrokerPin, OUTPUT);
	digitalWrite(BrokerPin, LOW);
	pinMode(SensorPin, OUTPUT);
	digitalWrite(SensorPin, LOW);
	Serial.begin(9600);
	TimeSet();
	if (Ethernet.begin(mac))
	{
		digitalWrite(NetworkPin, HIGH);
	}
	Serial.print("Starting Setup Routines, network is at: ");
	Serial.println(Ethernet.localIP());
	connectSDCard();
	if (hasSDCard == true)
	{
		readConfig();
	}
	connectMQTT();
	connectSensor();
	digitalWrite(MeasurePin, LOW);
	Serial.println("Ready.");
}




void loop() {
	if (!client.isConnected())
	{
		digitalWrite(BrokerPin, LOW);

		Serial.println("mqtt reconnect");
		connectMQTT();
	}


	// Der MQTT-Client muss gelegentlich aufgerufen werden, um Nachrichten empfangen zu k�nnen. Wert in ms
	int rc = client.yield(10);
	delay(100);

}

