// 
// 
// 

#include "updater.h"
#include <Arduino.h>
#include "FlashTxx.h"		// TLC/T3x/T4x/TMM flash primitives

void update_firmware(String FileName)
{
	uint32_t buffer_addr;
	uint32_t buffer_size;

	if (firmware_buffer_init(&buffer_addr, &buffer_size) == 0) {
		Serial.println("unable to create buffer\n");
		return;
	}

	static char line[96];					// buffer for hex lines
	static char data[32] __attribute__((aligned(8)));	// buffer for hex data
	hex_info_t hex = {					// intel hex info struct
	  data, 0, 0, 0,					//   data,addr,num,code
	  0, 0xFFFFFFFF, 0, 					//   base,min,max,
	  0, 0						//   eof,lines
	};

	Serial.println("waiting for file...");

	Serial.println("waiting for hex lines...");
	File  dataFile = SD.open(FileName.c_str(), FILE_READ);
	if (!dataFile)
	{
		Serial.print("could not open file: ");
		Serial.println(FileName.c_str());

	}
	else
	{
		dataFile.seek(0);
	}
	File* pdataFile = &dataFile;

	// read and process intel hex lines until EOF or error

	while (!hex.eof) {
		read_ascii_line(pdataFile, line, sizeof(line));
		Serial.println(line);
		if (parse_hex_line((const char*)line, hex.data, &hex.addr, &hex.num, &hex.code) == 0) {
			Serial.println("abort - bad hex line");
			return;
		}
		else if (process_hex_record(&hex) != 0) { // error on bad hex code
			Serial.println("abort - invalid hex code");// , hex.code);
			return;
		}
		else if (hex.code == 0) { // if data record
			uint32_t addr = buffer_addr + hex.base + hex.addr - FLASH_BASE_ADDR;
			if (hex.max > (FLASH_BASE_ADDR + buffer_size)) {
				Serial.println("abort - max address too large"); // , hex.max);
				return;
			}
			else if (!IN_FLASH(buffer_addr)) {
				Serial.println("writing to memory...");
				memcpy((void*)addr, (void*)hex.data, hex.num);
			}
			else if (IN_FLASH(buffer_addr)) {
				Serial.println("writing to flash...");
				int error = flash_write_block(addr, hex.data, hex.num);
				if (error) {
					Serial.print("abort - error ");
					Serial.print(error);
					Serial.println(" in flash_write_block()");
					return;
				}
			}
		}
		hex.lines++;
	}

	Serial.println("hex file read");//, hex.lines, hex.max - hex.min, hex.min, hex.max);
	// check FSEC value in new code -- abort if incorrect
#if defined(KINETISK) || defined(KINETISL)
	uint32_t value = *(uint32_t*)(0x40C + buffer_addr);
	if (value == 0xfffff9de) {
		serial->printf("new code contains correct FSEC value %08lX");
	}
	else {
		serial->printf("abort - FSEC value %08lX should be FFFFF9DE");
		return;
}
#endif

	// check FLASH_ID in new code - abort if not found
	if (check_flash_id(buffer_addr, hex.max - hex.min)) {
		Serial.println("new code contains correct target ID %s");
	}
	else {
		Serial.println("abort - new code missing string %s");
		return;
	}


	// do not burn yet
	return;
	// move new program from buffer to flash, free buffer, and reboot
	flash_move(FLASH_BASE_ADDR, buffer_addr, hex.max - hex.min);

	// should not return from flash_move(), but put REBOOT here as reminder
	REBOOT;
}



//******************************************************************************
// read_ascii_line()	read ascii characters until '\n', '\r', or max bytes
//******************************************************************************
void read_ascii_line(File* pmyfile, char* line, int maxbytes)
{
	File myfile = *pmyfile;
	int c;
	int nchar;
	nchar = 0;
	c = 0;
	//Serial.println(myfile.position());


	while (nchar < maxbytes && !(c == '\n' || c == '\r')) {
		c = myfile.read();
		//Serial.println(myfile.position());
		line[nchar++] = c;
		//Serial.println(line);
	}
	line[nchar - 1] = 0;	// null-terminate
	auto d = myfile.peek();
	if (d == '\n' || d == '\r') 
	{
		auto dmy = myfile.read();
		Serial.println(myfile.position());
	}
}

//******************************************************************************
// process_hex_record()		process record and return okay (0) or error (1)
//******************************************************************************
int process_hex_record(hex_info_t* hex)
{
	if (hex->code == 0) { // data -- update min/max address so far
		if (hex->base + hex->addr + hex->num > hex->max)
			hex->max = hex->base + hex->addr + hex->num;
		if (hex->base + hex->addr < hex->min)
			hex->min = hex->base + hex->addr;
	}
	else if (hex->code == 1) { // EOF (:flash command not received yet)
		hex->eof = 1;
	}
	else if (hex->code == 2) { // extended segment address (top 16 of 24-bit addr)
		hex->base = ((hex->data[0] << 8) | hex->data[1]) << 4;
	}
	else if (hex->code == 3) { // start segment address (80x86 real mode only)
		return 1;
	}
	else if (hex->code == 4) { // extended linear address (top 16 of 32-bit addr)
		hex->base = ((hex->data[0] << 8) | hex->data[1]) << 16;
	}
	else if (hex->code == 5) { // start linear address (32-bit big endian addr)
		hex->base = (hex->data[0] << 24) | (hex->data[1] << 16)
			| (hex->data[2] << 8) | (hex->data[3] << 0);
	}
	else {
		return 1;
	}

	return 0;
}

//******************************************************************************
// Intel Hex record foramt:
//
// Start code:  one character, ASCII colon ':'.
// Byte count:  two hex digits, number of bytes (hex digit pairs) in data field.
// Address:     four hex digits
// Record type: two hex digits, 00 to 05, defining the meaning of the data field.
// Data:        n bytes of data represented by 2n hex digits.
// Checksum:    two hex digits, computed value used to verify record has no errors.
//
// Examples:
//  :10 9D30 00 711F0000AD38000005390000F5460000 35
//  :04 9D40 00 01480000 D6
//  :00 0000 01 FF
//******************************************************************************

/* Intel HEX read/write functions, Paul Stoffregen, paul@ece.orst.edu */
/* This code is in the public domain.  Please retain my name and */
/* email address in distributed copies, and let me know about any bugs */

/* I, Paul Stoffregen, give no warranty, expressed or implied for */
/* this software and/or documentation provided, including, without */
/* limitation, warranty of merchantability and fitness for a */
/* particular purpose. */

// type modifications by Jon Zeeff

/* parses a line of intel hex code, stores the data in bytes[] */
/* and the beginning address in addr, and returns a 1 if the */
/* line was valid, or a 0 if an error occured.  The variable */
/* num gets the number of bytes that were stored into bytes[] */

#include <stdio.h>		// sscanf(), etc.
#include <string.h>		// strlen(), etc.

int parse_hex_line(const char* theline, char* bytes,
	unsigned int* addr, unsigned int* num, unsigned int* code)
{
	unsigned sum, len, cksum;
	const char* ptr;
	int temp;

	*num = 0;
	if (theline[0] != ':')
		return 0;
	if (strlen(theline) < 11)
		return 0;
	ptr = theline + 1;
	if (!sscanf(ptr, "%02x", &len))
		return 0;
	ptr += 2;
	if (strlen(theline) < (11 + (len * 2)))
		return 0;
	if (!sscanf(ptr, "%04x", (unsigned int*)addr))
		return 0;
	ptr += 4;
	/* Serial.printf("Line: length=%d Addr=%d\n", len, *addr); */
	if (!sscanf(ptr, "%02x", code))
		return 0;
	ptr += 2;
	sum = (len & 255) + ((*addr >> 8) & 255) + (*addr & 255) + (*code & 255);
	while (*num != len)
	{
		if (!sscanf(ptr, "%02x", &temp))
			return 0;
		bytes[*num] = temp;
		ptr += 2;
		sum += bytes[*num] & 255;
		(*num)++;
		if (*num >= 256)
			return 0;
	}
	if (!sscanf(ptr, "%02x", &cksum))
		return 0;

	if (((sum & 255) + (cksum & 255)) & 255)
		return 0;     /* checksum error */
	return 1;
}
